/*************************************************************************/
/*                                                                       */
/* MODULE NAME        DDESERV.C                                          */
/*                                                                       */
/* DESCRIPTIVE NAME   DDE Server for CICS OS/2 ECI                       */
/*                                                                       */
/* Statement:         Sample Code - No Warranty Provided.                */
/*                                                                       */
/*                    (c) Copyright IBM Corp. 1994.                      */
/*                                                                       */
/*                    All rights reserved.                               */
/*                                                                       */
/* Status:            Version 1 Release 0 Modification 1                 */
/*                                                                       */
/*  NOTES :-                                                             */
/*    DEPENDENCIES = For Windows 3.x                                     */
/*    RESTRICTIONS = none                                                */
/*    MODULE TYPE  = Source Module                                       */
/*    PROCESSOR    = PS/2 and PC                                         */
/*                                                                       */
/*************************************************************************/
/*                                                                       */
/* Change History:                                                       */
/*    21/12/94:    Initial Module Generation                             */
/*    23/01/95:    Fix to allow Word for Windows Macro to work.   -KJT1  */
/*                                                                       */
/*                                                                       */
/*                                                                       */
/*                                                                       */
/*************************************************************************/
/* PROGRAM NAME: DDESERV                                                 */
/*                                                                       */
/* Sample of DDE Server for CICS OS/2 ECI                                */
/* Compilable for Windows 3.x                                            */
/*                                                                       */
/*************************************************************************/

/* Windows include files: */
#include <windows.h>
#include <dde.h>

/* CICS OS/2 ECI include file: */
#include "faaecih.h"

/* Program include file: */
#include "ddeserv.h"

/* C include files: */
#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <stdlib.h>

/***************************************************************************/
/*                                                                         */
/*  FUNCTION: WinMain                                                      */
/*                                                                         */
/*  PURPOSE:  Initialises the main windows and starts the message loop     */
/*                                                                         */
/***************************************************************************/
int PASCAL WinMain ( HINSTANCE hInst,
                     HINSTANCE hPrevInst,
                     LPSTR     lpCmd,
                     int       nCmdShow )
{
  MSG          msg  ;
  WNDCLASS     wc   ;
  RECT         rc   ;

  /* Save the instance handle in a global */
  hInstance = hInst ;

  /* Initialise the time out constant */
  nAckTimeOut = DEFAULT_ACK_TIME_OUT_MILLISEC ;

  /* Application specific initialisation */
  if ( hPrevInst == (HINSTANCE) NULL ) {

     /* Create the window classes -  */
     wc.style         = CS_HREDRAW | CS_VREDRAW ;
     wc.lpfnWndProc   = MainWindowProc ;
     wc.cbClsExtra    = 0 ;
     wc.cbWndExtra    = 0 ;
     wc.hInstance     = hInstance ;
     wc.hIcon         = LoadIcon(NULL, IDI_APPLICATION) ;
     wc.hCursor       = LoadCursor(NULL, IDC_ARROW) ;
     wc.hbrBackground = (HBRUSH) COLOR_WINDOW+1 ;
     wc.lpszMenuName  = ECI_MENU ;
     wc.lpszClassName = ECI_CLASS ;

     if (!RegisterClass (&wc)) {
       MessageBox ( NULL, "Initialization Failure 1", APPL_NAME, MB_ICONSTOP|MB_OK ) ;
       return FALSE ;
     } /* endif */

     wc.style         = NULL ;
     wc.lpfnWndProc   = DDEWindowProc ;
     wc.cbClsExtra    = 0 ;
     wc.cbWndExtra    = 0 ;
     wc.hInstance     = hInstance ;
     wc.hIcon         = NULL ;
     wc.hCursor       = NULL ;
     wc.hbrBackground = NULL ;
     wc.lpszMenuName  = NULL ;
     wc.lpszClassName = DDESERVER_CLASS ;

     if (!RegisterClass (&wc)) {
        MessageBox ( NULL, "Initialization Failure 2", APPL_NAME, MB_ICONSTOP|MB_OK ) ;
        return FALSE ;
     } /* end if */
  } /* end if */
  /* check to see if appl is already running and if so, return */

  if ( !FindWindow ( ECI_CLASS,  APPL_NAME ) == NULL ) {
     MessageBox ( NULL, "Application already started", APPL_NAME, MB_ICONSTOP|MB_OK ) ;
     return FALSE ;
  } /* end if */

  /* Create the window */

  hwndMain = CreateWindow ( ECI_CLASS,
                            APPL_NAME,
                            WS_OVERLAPPEDWINDOW | WS_CLIPCHILDREN,
                            25, 25, 500, 300,
                            NULL,
                            NULL,
                            hInstance,
                            NULL ) ;

  /* Check the result: */

  if ( hwndMain == NULL ) {
     MessageBox ( NULL, "Initialization Failure 3", APPL_NAME, MB_ICONSTOP|MB_OK ) ;
     return FALSE ;
  } /* endif */

  /* Create a listbox as the client area */

  GetClientRect ( hwndMain, &rc ) ;

  hwndList = CreateWindow ( "Listbox",
                             NULL,
                             WS_CHILD | WS_VISIBLE | WS_VSCROLL |
                             LBS_NOTIFY | LBS_NOINTEGRALHEIGHT,
                             rc.left, rc.top,
                             rc.right - rc.left, rc.bottom - rc.top,
                             hwndMain,
                             NULL,
                             hInstance,
                             NULL ) ;
  /* Check the result */

  if ( hwndList == NULL ) {
     MessageBox ( NULL, "Initialization Failure 4", APPL_NAME, MB_ICONSTOP|MB_OK ) ;
     return FALSE ;
  } /* endif */

  /* Show the window */

  ShowWindow   ( hwndMain, nCmdShow ) ;
  UpdateWindow ( hwndMain ) ;

  /* Do the standard message loop */

  while ( GetMessage ( &msg, NULL, NULL, NULL ) ) {
    TranslateMessage ( &msg ) ;
    DispatchMessage  ( &msg ) ;
  } /* end while */

  /* End the program */

  return msg.wParam ;

}

/***************************************************************************/
/*                                                                         */
/*  FUNCTION: MainWindowProc                                               */
/*                                                                         */
/*  PURPOSE:  The main window procedure - handles all top level messages   */
/*                                                                         */
/***************************************************************************/
long FAR PASCAL _export MainWindowProc ( HWND     hwnd,
                                         UINT     msg,
                                         WPARAM   wParam,
                                         LPARAM   lParam )
{
  short           sRetCode = 0 ;
  long            lRetCode = 0L ;
  FARPROC         lpAboutDlgProc ;


  switch ( msg ) {

    case WM_SIZE :
      MoveWindow ( hwndList,
                   0, 0,
                   LOWORD ( lParam ),
                   HIWORD ( lParam ),
                   TRUE ) ;
      break ;

    case WM_COMMAND :
      switch ( wParam ) {


        case ECI_ITEM_EXIT :
          PostMessage ( hwnd, WM_CLOSE, 0, 0L ) ;
          break ;

        case ECI_ITEM_CLEAR :
          SendMessage ( hwndList, LB_RESETCONTENT, 0, 0L ) ;
          break ;

        case ECI_ITEM_CALL :

          /* set up the commarea to be sent and call the ECI */
          strcpy ( szCommArea, "Windows Sample ECI Call" ) ;
          sRetCode = ECICall ( "ECITEST ", hwndMain ) ;
          break ;

        case ECI_ITEM_ABOUT :

          lpAboutDlgProc = MakeProcInstance((FARPROC)AboutDlgProc, hInstance) ;
          DialogBox( hInstance,
                     ECI_ABOUT,
                     hwndMain,
                     (DLGPROC) lpAboutDlgProc) ;
                     FreeProcInstance(lpAboutDlgProc) ;
          break;

        default :
        break ;

      } /* end switch(wParam) */

    break;   /* case WM_COMMAND */

    case ECI_MSG :
      /* display the ECI return code */
      Response ( "ECI Async Notify Msg ", (short) wParam ) ;

      if (ECIRequestData() != ECI_NO_ERROR) {
        /* The ECI call has failed. Write an error message to screen */
        strcpy( szCommArea , "ECI Failure, check DDESERV") ;
      } /* end if */

      Info ( "Commarea returned is '%s'",  szCommArea ) ;

      break ;

    case WM_DDE_INITIATE :

      Info ( "DDE_INITIATE received") ;
      ServerInitiate((HWND) wParam, lParam) ;

      break ;

    case WM_DESTROY :

      /* Terminate all DDE conversations before destroying client window */
      TerminateConversations() ;
      PostQuitMessage(0) ;

      break ;

    default :

      return DefWindowProc ( hwnd, msg, wParam, lParam ) ;

  } /* end switch(msg) */

  return lRetCode ;

}  /* end function MainWindowProc */


/**************************************************************************/
/*                                                                        */
/*  FUNCTION: ECICall                                                     */
/*                                                                        */
/*  PURPOSE:  Call ECI requesting ASYNC_NOTIFY passing commarea           */
/*                                                                        */
/**************************************************************************/

short ECICall (char szProgram[TOPIC_MAX_SIZE+1], HWND hwndToNotify )
{

  short   sSize ;
  short   sRetCode ;

  /* for this sample - commarea is limited to 256 bytes */

  sSize = sizeof ( szCommArea ) ;
  Info ( "Calling ECI - sending %d bytes. Commarea is '%s'.  Window to Notify is %d.",
         strlen(szCommArea),  szCommArea, hwndToNotify ) ;

  /* clear the ECI control block */
  memset ( &eci_parms, 0, sizeof ( ECI_PARMS ) ) ;

  /* set up the ECI control block for this call */
  eci_parms.eci_call_type                         = ECI_ASYNC_NOTIFY_MSG ;
  memcpy ( &eci_parms.eci_program_name, szProgram , 8 ) ;
  memcpy ( &eci_parms.eci_userid      , "SYSAD   ", 8 ) ;
  memcpy ( &eci_parms.eci_password    , "SYSAD   ", 8 ) ;
  eci_parms.eci_commarea                          = szCommArea ;
  eci_parms.eci_commarea_length                   = sSize ;
  eci_parms.eci_extend_mode                       = ECI_NO_EXTEND ;
  eci_parms.eci_async_notify.win_fields.hwnd      = (unsigned short) hwndToNotify ;
  eci_parms.eci_async_notify.win_fields.hinstance = (unsigned short) hInstance ;
  eci_parms.eci_message_id                        = ECI_MSG ;
  eci_parms.eci_luw_token                         = ECI_LUW_NEW ;

  /*  and call the ECI */
  sRetCode = FaaExternalCall ( &eci_parms ) ;
  Response ( "ECI ", sRetCode) ;

  return sRetCode ;

} /* end function ECICall */


/**************************************************************************/
/*                                                                        */
/*  FUNCTION: ECIRequestData                                              */
/*                                                                        */
/*  PURPOSE:  Call ECI to retrieve data after ECI has posted ASYNC_NOTIFY */
/*                                                                        */
/**************************************************************************/

short ECIRequestData (void)
{

  short   sSize ;
  short   sRetCode ;

  Info ("Calling ECI to obtain commarea after ASYNC_NOTIFY.") ;

  /* clear the commarea */
  sSize = sizeof ( szCommArea ) ;
  memset ( szCommArea , 0 , sSize ) ;

  /* clear the ECI control block */
  memset ( &eci_parms,  0,  sizeof ( ECI_PARMS ) ) ;

  eci_parms.eci_call_type                   = ECI_GET_REPLY ;
  eci_parms.eci_commarea                    = szCommArea ;
  eci_parms.eci_commarea_length             = sSize ;

  /* call the ECI */
  sRetCode = FaaExternalCall ( &eci_parms ) ;
  /* display the return code */
  Response ( "ECI ", sRetCode ) ;

  return sRetCode ;

} /* end function ECIRequestData */


/***************************************************************************/
/*                                                                         */
/*  FUNCTION: DDEWindowProc                                                */
/*                                                                         */
/*  PURPOSE:  Handles all DDE messages from client window                  */
/*                                                                         */
/***************************************************************************/

long FAR PASCAL __export DDEWindowProc( HWND     hwnd,
                                        UINT     msg,
                                        WPARAM   wParam,
                                        LPARAM   lParam )

{

  long                lRetCode = 0L ;

  switch (msg)  {

    case WM_DDE_TERMINATE :
      ServerTerminate(hwnd, (HWND)wParam) ;
      break ;

    case WM_DDE_REQUEST :
      ServerRequest(hwnd, (HWND)wParam,  lParam) ;
      break ;

    case ECI_MSG :
      /* display the ECI return code */
      Response ( "ECI Async Notify Msg ", (short) wParam ) ;

      if (ECIRequestData() != ECI_NO_ERROR) {
        /* the ECI call has failed. Copy an error message into the commarea */
        strcpy( szCommArea , "ECI Failure, check DDESERV") ;
      } /* end if */

      Info ( "Commarea returned is  '%s'",  szCommArea ) ;

      /* now we need to reestablish Conversation details to send the data */
      /* back to the requesting client. Get the item atom and extract the */
      /* string from the atom table                                       */

      GlobalGetAtomName(GetConvItem(hwnd), szItem, ITEM_NAME_MAX_SIZE) ;

      /* the return value goes into szItemValue */
      strcpy( szItemValue, szCommArea ) ;

      /* send data now and don't ask for acknowledgement */

      SendData(hwnd, GetHwndDDEClient(hwnd), szItem, szItemValue, TRUE ) ;

      /* delete the atom representing the item sent by the client as  */
      /* SendData will create a new atom to send back to the client.  */

      GlobalDeleteAtom(HIWORD(lParam)) ;         
      
      /* and delete the topic atom here aswell */                          // kjt1
      GlobalDeleteAtom( GetConvTopic( hwnd )) ;                            // kjt1

      
      break ;

    default :

      return (DefWindowProc(hwnd, msg, wParam, lParam)) ;

    }  /* end switch */

    return lRetCode ;

}   /* end function DDEWindowProc */

/***************************************************************************/
/*                                                                         */
/*  FUNCTION: ServerInitiate                                               */
/*                                                                         */
/*  PURPOSE:  To initialise child window and send acknowledgement if       */
/*            application and topic are for this application               */
/*                                                                         */
/***************************************************************************/
void ServerInitiate( HWND hwndDDEClient, LONG lParam )
{
  HWND  hwndDDEServer ;
  ATOM  atomApplicationRcvd ;
  ATOM  atomTopicRcvd ;
  ATOM  atomApplicationReturn ;
  ATOM  atomTopicReturn ;
  char  szApplication [APP_MAX_SIZE+1] ;
  char  szTopic       [TOPIC_MAX_SIZE+1] ;

  /* get the application name and test to see if it matches our name */

  atomApplicationRcvd = LOWORD(lParam) ;
  if ( atomApplicationRcvd) {
    GlobalGetAtomName(atomApplicationRcvd, szApplication, sizeof(szApplication)) ;
  } /* end if */

  if (atomApplicationRcvd && strcmpi(szApplication,APPL_NAME)) {
    /* if application was specified but it wasn't for us */
    Info ("Application Not DDESERV. Name = %s", szApplication) ;
    return ;
  }  /* end if */

  /* get the topic name and save it globally */

  atomTopicRcvd = HIWORD(lParam) ;
  if ( atomTopicRcvd ) {
    GlobalGetAtomName(atomTopicRcvd, szTopic, sizeof(szTopic)) ;
    Info ("Application is '%s', Topic is '%s'.", szApplication, szTopic) ;
  }  /* end if */

  /* Application and Topic are for us so create the child window */
  /* which will handle the conversation                          */

  hwndDDEServer = CreateWindow(
             DDESERVER_CLASS,
             APPL_NAME,
             WS_CHILD,               /* not visible */
             0, 0, 0, 0,             /* no position or dimensions */
             hwndMain,               /* parent */
             NULL,                   /* no menu */
             hInstance,
             NULL) ;

  if ( hwndDDEServer == (HWND) NULL ) {
    MessageBox ( NULL, "DDE Initialization failure", APPL_NAME, MB_ICONSTOP|MB_OK ) ;
    return ;
  } /* endif */

  Info ("DDE Server window initialised OK, handle is %d" , hwndDDEServer ) ;

  /* add the conversation to the list and reply to client with ACK */

  if (AddConv(hwndDDEServer, hwndDDEClient )) {

    /* create new atoms to return to client as client should delete existing atoms */
    atomApplicationReturn = GlobalAddAtom(szApplication) ;
    atomTopicReturn = GlobalAddAtom(szTopic) ;
    
    strcpy( szTopicName , szTopic ) ;     /* save topic name globally */
    
    SendMessage(hwndDDEClient,
                WM_DDE_ACK,
                (WPARAM)hwndDDEServer,
                MAKELONG(atomApplicationReturn, atomTopicReturn)) ;

    Info ("Sent DDE_ACK to client %d." ,  hwndDDEClient ) ;

    /* save the topic atom in the conversation structure */
    AddConvTopic ( hwndDDEServer, atomTopicReturn ) ;
    /* delete the application atom (decrement reference count) */
    /* We leave the topic atom in the global table as we need this later */ 
    GlobalDeleteAtom ( atomApplicationReturn) ;

  } /* end if  */

  else {
  /* AddConv failed - at maximum conversation count - destroy the window */
  /* client will timeout waiting for ACK */
    DestroyWindow(hwndDDEServer) ;
    Info ("Max No of clients attached. DDE Server window %d destroyed.", hwndDDEServer) ;
  } /* end else  */
  return ;

} /* end function ServerInitiate */

/**************************************************************************/
/*                                                                        */
/*  FUNCTION: ServerTerminate                                             */
/*                                                                        */
/*  PURPOSE:  To handle termination of DDE conversation                   */
/*                                                                        */
/**************************************************************************/
void ServerTerminate(HWND hwndDDEServer, HWND hwndDDEClient)
{
  Info ("DDE_TERMINATE received from client %d.",  hwndDDEClient ) ;

  if (! IsConvInTerminateState( hwndDDEClient )) {
    /* Client has requested terminate: responsd with terminate */
    PostMessage(hwndDDEClient, WM_DDE_TERMINATE, (WPARAM)hwndDDEServer, 0L) ;
  }  /* end if */

  RemoveConv   (hwndDDEServer) ;
  DestroyWindow(hwndDDEServer) ;
  Info ("DDE Server window %d destroyed.", hwndDDEServer) ;
  return ;
}  /* end function ServerTerminate */

/**************************************************************************/
/*                                                                        */
/*  FUNCTION: ServerRequest                                               */
/*                                                                        */
/*  PURPOSE:  To handle DDE_REQUEST message                               */
/*                                                                        */
/**************************************************************************/
void ServerRequest(HWND hwndDDEServer, HWND hwndDDEClient, LONG lParam)
{
                                                
  int     sTopic ;                                             
  short   sRetCode ;
  char    szTopic[TOPIC_MAX_SIZE+1] ;          /* to store topic name */

  Info ("DDE_REQUEST received from client %d." ,  hwndDDEClient ) ;

  /* Get the string referenced by the atom passed on the DDE_REQUEST.*/
  /* This corresponds to the 'item' requested and will be passed as  */
  /* a commarea to CICS OS/2 via the ECI                             */

  GlobalGetAtomName(HIWORD(lParam), szItem, sizeof(szItem)) ;
  
  /* check that the 'item' is only a simple text string */

  if (LOWORD(lParam) != CF_TEXT) {

    /* this app supports only CF_TEXT, reject the request with  */
    /* a NACK response and reusing the original atom item       */

    PostMessage( hwndDDEClient,
                 WM_DDE_ACK,
                 (WPARAM)hwndDDEServer,
                 MAKELONG(0, HIWORD(lParam))) ;
    return ;
  } /* end if */


  /* re-establish the topic for the conversation using the atom stored */
  /* in the conversation structure to get the topic text string        */

  sTopic = GlobalGetAtomName((GetConvTopic( hwndDDEServer )), szTopic, sizeof(szTopic)) ;
  
  if ( sTopic == 0 ) {
     /* client appl has deleted the global atom for the topic */             // kjt1
     /* Use the globally stored topic for demo purposes only  */             // kjt1
     
     strcpy ( szTopic, szTopicName ) ;                                       // kjt1
  }  /* end if */   
  

  /* Now do the ECI call using the 'topic' as the program name */
  /* and the 'item' as the contents of a commarea to be passed */

  strcpy ( szCommArea, szItem ) ;

  sRetCode = ECICall ( szTopic, hwndDDEServer ) ;

  if ( sRetCode != ECI_NO_ERROR ) {

    /* the ECI call has failed. Copy an error message into the */
    /* return value and send with as a normal DDE response     */

    strcpy( szItemValue, "ECI Failure. Check Server.") ;
    SendData(hwndDDEServer, hwndDDEClient, szItem, szItemValue, TRUE) ;

    /* delete the atom representing the item sent by the client as  */
    /* SendData will create a new atom to send back to the client.  */

    GlobalDeleteAtom(HIWORD(lParam));
    
    /* and delete the topic atom here aswell */                          // kjt1
    GlobalDeleteAtom( GetConvTopic( hwndDDEServer)) ;                    // kjt1
    
  }  /* end if */

  /* ECI Call will post a message when it returns from CICS OS/2.   */
  /* to allow us to reply to client we must save the atom which     */
  /* represents the requested item in the conv structure so we can  */
  /* restore it when the ECI_MSG is received.                       */

  AddConvItem ( hwndDDEServer, HIWORD(lParam)) ;

  /* so now we just return and wait for ECI_MSG to be driven   */

  return ;

}   /* end function ServerRequest */


/**************************************************************************/
/*                                                                        */
/*  FUNCTION: SendData                                                    */
/*                                                                        */
/*  PURPOSE:  Send DDE data to specified client                           */
/*                                                                        */
/**************************************************************************/
void SendData( HWND hwndDDEServer,
               HWND hwndDDEClient,
               char * szItemName,
               char * szItemValue,
               BOOL bRequestData )
{
  ATOM          atomItem ;
  HANDLE        hData ;
  DDEDATA FAR * lpData ;

  /* Allocate memory to the size of the DDE data header, plus the data: */

  hData = GlobalAlloc(GMEM_MOVEABLE | GMEM_DDESHARE,
                      (LONG)sizeof(DDEDATA) +lstrlen(szItemValue) + 2) ;
  if ( hData == NULL ) {
    MessageBox ( NULL, "Unable to Allocate DDE Control block. Reply failure",
                 APPL_NAME, MB_ICONEXCLAMATION|MB_OK ) ;
    return ;
  } /* end if */

  lpData = (DDEDATA FAR*)GlobalLock(hData) ;
  if ( lpData == NULL ) {
    MessageBox ( NULL, "Unable to lock DDE Control block. Reply failure",
                 APPL_NAME, MB_ICONEXCLAMATION|MB_OK ) ;
    GlobalFree(hData) ;
    return ;
  } /* end if */

  lpData->fAckReq   = FALSE ;        /* sent to acknowledge DDE_REQUEST */
  lpData->cfFormat  = CF_TEXT ;      /* string <CR> <LF> only */
  lpData->fRelease  = TRUE ;         /* client appl is to release the global memory */
  if ( bRequestData ) {
    lpData->fResponse = 1 ;          /* We do require DDE_ACK response */
  }
  else {
    lpData->fResponse = 0 ;          /* We don't require DDE_ACK response */
  } /* end if */

  lstrcpy((LPSTR)lpData->Value, (LPSTR)szItemValue) ;

  /* Note: each line of CF_TEXT data is terminated by <CR> <LF> */

  lstrcat((LPSTR)lpData->Value, (LPSTR)"\r\n") ;
  GlobalUnlock(hData) ;
  atomItem = GlobalAddAtom((LPSTR)szItemName) ;

  if (!PostMessage(hwndDDEClient,
                   WM_DDE_DATA,
                   (WPARAM)hwndDDEServer,
                   MAKELONG(hData, atomItem))) {
    /* if the PostMessage call fails we must delete the global memory */
    /* and the atom which we previously created. If it succeeds then  */
    /* the client appl should do this                                 */

    GlobalFree(hData) ;
    GlobalDeleteAtom(atomItem) ;
    return ;
  } /* end if */

}   /* end function SendData */


/**************************************************************************/
/*                                                                        */
/*  FUNCTION: Terminate Converations                                      */
/*                                                                        */
/*  PURPOSE:  After WM_DESTROY, we must terminate all DDE converations    */
/*                                                                        */
/**************************************************************************/
void TerminateConversations( void )
{
  HWND  hwndDDEServer ;
  LONG  lTimeOut ;
  MSG   msg ;

   /* Send DDE_TERMINATE to each client in conversation with us */
   hwndDDEServer = NULL ;
   while (hwndDDEServer = GetNextConv(hwndDDEServer)) {
      SendTerminate(hwndDDEServer, GetHwndDDEClient(hwndDDEServer)) ;
   } /* end while */

   /* Wait for all conversations to terminate OR for time out */
   lTimeOut = GetTickCount() + (LONG)nAckTimeOut ;

   while (PeekMessage(&msg, NULL, WM_DDE_FIRST, WM_DDE_LAST, PM_REMOVE)) {

     DispatchMessage (&msg) ;
     if (msg.message == WM_DDE_TERMINATE) {
       if (!AtLeastOneConvActive())
       break ;
     } /* end if */

     if (GetTickCount() > (DWORD)lTimeOut)
       break ;
   } /* end while */

   return ;

} /* end function TerminateConversations */


/**************************************************************************/
/*                                                                        */
/*  FUNCTION: AddConv                                                     */
/*                                                                        */
/*  PURPOSE:  Register a new conversation with a client                   */
/*                                                                        */
/**************************************************************************/
BOOL AddConv(HWND hwndDDEServer, HWND hwndDDEClient )
{
  struct DDECONV * pConv ;

  if (nConvCount >= CONV_MAX_COUNT) {
    return (FALSE) ;
  } /* end if */

  if (FindConv(hwndDDEServer) != NULL) {
    /* conversation already added */
    return (FALSE) ;
  } /* end if */

  pConv = Conv + nConvCount++ ;
  pConv->hwndDDEServer = hwndDDEServer ;
  pConv->hwndDDEClient = hwndDDEClient ;
  pConv->aClientTopic = (ATOM) NULL ;
  pConv->aClientItem = (ATOM) NULL ;
  pConv->bInClientRequestedTerminate = FALSE ;
  return (TRUE) ;

} /* end function AddConv */


/**************************************************************************/
/*                                                                        */
/*  FUNCTION: AddConvTopic                                                */
/*                                                                        */
/*  PURPOSE:  Update the conversation to contain atom for Topic           */
/*                                                                        */
/**************************************************************************/
BOOL AddConvTopic ( HWND hwndDDEServer, ATOM aClientTopic )
{

  struct DDECONV * pConv ;

  if (!(pConv = FindConv(hwndDDEServer)))
    return FALSE ;


  pConv->aClientTopic = aClientTopic ;
  return TRUE ;

} /* end function AddConvTopic */


/**************************************************************************/
/*                                                                        */
/*  FUNCTION: AddConvItem                                                 */
/*                                                                        */
/*  PURPOSE:  Update the conversation to contain atom for Topic           */
/*                                                                        */
/**************************************************************************/
BOOL AddConvItem ( HWND hwndDDEServer, ATOM aClientItem )
{

  struct DDECONV * pConv ;

  if (!(pConv = FindConv(hwndDDEServer)))
    return FALSE ;


  pConv->aClientItem = aClientItem ;
  return TRUE ;

} /* end function AddConvItem */

/**************************************************************************/
/*                                                                        */
/*  FUNCTION: GetConvTopic                                                */
/*                                                                        */
/*  PURPOSE:  Get the atom representing the 'Topic' of the conversation   */
/*            with the specified DDE Server Window                        */
/*                                                                        */
/**************************************************************************/
ATOM GetConvTopic ( HWND hwndDDEServer)
{
  struct DDECONV * pConv ;

  if (!(pConv = FindConv(hwndDDEServer)))
    return (ATOM) NULL ;
  return (pConv->aClientTopic) ;

} /* end function GetConvTopic */


/**************************************************************************/
/*                                                                        */
/*  FUNCTION: GetConvItem                                                 */
/*                                                                        */
/*  PURPOSE:  Get the atom representing the 'item' of the conversation    */
/*            with the specified DDE Server Window                        */
/*                                                                        */
/**************************************************************************/
ATOM GetConvItem ( HWND hwndDDEServer )
{
  struct DDECONV * pConv ;

  if (!(pConv = FindConv(hwndDDEServer)))
    return (ATOM) NULL ;
  return (pConv->aClientItem) ;

} /* end function GetConvItem */



/**************************************************************************/
/*                                                                        */
/*  FUNCTION: AtLeastoneConvActive                                        */
/*                                                                        */
/*  PURPOSE:  Test whether any conversations are still active             */
/*                                                                        */
/**************************************************************************/
BOOL AtLeastOneConvActive( void )
{
    return (nConvCount? TRUE: FALSE);
}   /* end function AtLeastOneConvActive */


/**************************************************************************/
/*                                                                        */
/*  FUNCTION: FindConv                                                    */
/*                                                                        */
/*  PURPOSE:  Find Conversation for specified DDE server window           */
/*                                                                        */
/**************************************************************************/
struct DDECONV * NEAR FindConv( HWND hwndDDEServer)
{
  struct DDECONV *  pConv ;
  int               nConvIndex ;

  for (nConvIndex = 0, pConv = Conv ;
       nConvIndex < nConvCount ;
       nConvIndex++, pConv++) {

    if (pConv->hwndDDEServer == hwndDDEServer)
      return (pConv) ;
  } /* end for */

  return (NULL) ;

}   /* end function FindConv */


/**************************************************************************/
/*                                                                        */
/*  FUNCTION: GetNextConv                                                 */
/*                                                                        */
/*  PURPOSE:  Get the next client handle from the list of conversations   */
/*                                                                        */
/**************************************************************************/
HWND GetNextConv(HWND hwndDDEServer)
{
  struct DDECONV * pConv ;
  int           nConvIndex ;

  if (hwndDDEServer) {
    for (nConvIndex = 0, pConv = Conv ;
         nConvIndex < nConvCount ;
         nConvIndex++, pConv++) {

      if (pConv->hwndDDEServer == hwndDDEServer) {

        if (++nConvIndex < nConvCount)
          return (++pConv)->hwndDDEServer ;
        else
          return (HWND) NULL ;
      } /* end if */
    }  /* end for */
    return (HWND) NULL ;
  } /* end if */

  if (nConvCount > 0)
    return (Conv[0].hwndDDEServer) ;
  else
    return (HWND) NULL ;

}   /* end function GetNextConv */


/**************************************************************************/
/*                                                                        */
/*  FUNCTION: RemoveConv                                                  */
/*                                                                        */
/*  PURPOSE:  Remove the conversation from the list                       */
/*                                                                        */
/**************************************************************************/
void RemoveConv( HWND hwndDDEServer)
{
  struct DDECONV   * pConv ;
  int              nConvIndex ;

  for (nConvIndex = 0, pConv = Conv ;
       nConvIndex < nConvCount ;
       nConvIndex++, pConv++) {

    if (pConv->hwndDDEServer == hwndDDEServer)
      break ;
  } /* end for */

  nConvCount-- ;
  while (nConvIndex < nConvCount) {

    *pConv = *(pConv+1) ;
    nConvIndex++ ;
    pConv++ ;
  }  /* end while */

  return;

} /* end function RemoveConv */


/**************************************************************************/
/*                                                                        */
/*  FUNCTION: GetHwndDDEClient                                            */
/*                                                                        */
/*  PURPOSE:  Get the hwnd of the client in conversation with the         */
/*            specified DDE Server Window                                 */
/*                                                                        */
/**************************************************************************/
HWND GetHwndDDEClient( HWND hwndDDEServer )
{
  struct DDECONV * pConv ;

  if (!(pConv = FindConv(hwndDDEServer)))
    return (HWND) NULL ;
  return (pConv->hwndDDEClient) ;

} /* end function GetHwndDDEClient */


/**************************************************************************/
/*                                                                        */
/*  FUNCTION: SendTerminate                                               */
/*                                                                        */
/*  PURPOSE:  Post terminate message and set the terminating flag for     */
/*            the conversation                                            */
/*                                                                        */
/**************************************************************************/
void SendTerminate( HWND hwndDDEServer, HWND hwndDDEClient )
{

  SetConvInTerminateState(hwndDDEServer) ;
  PostMessage(hwndDDEClient, WM_DDE_TERMINATE, (WPARAM)hwndDDEServer, 0L) ;
  return ;

}

/**************************************************************************/
/*                                                                        */
/*  FUNCTION: SetConvInTerminateState                                     */
/*                                                                        */
/*  PURPOSE:  Set the specified conversations terminate state to true     */
/*                                                                        */
/**************************************************************************/
void SetConvInTerminateState( HWND hwndDDEServer )
{
  struct DDECONV * pConv ;

  if (pConv = FindConv(hwndDDEServer))
      pConv->bInClientRequestedTerminate = TRUE ;
  return;

} /* end function SetConvInTerminateState */


/**************************************************************************/
/*                                                                        */
/*  FUNCTION: IsConvInTerminateState                                      */
/*                                                                        */
/*  PURPOSE:  Test whether conversation with specified client is in       */
/*            process of being terminated                                 */
/*                                                                        */
/**************************************************************************/
BOOL IsConvInTerminateState( HWND hwndDDEServer )
{
  struct DDECONV * pConv;

  if (pConv = FindConv(hwndDDEServer))
    return (pConv->bInClientRequestedTerminate) ;
  else
    return (FALSE);

}   /* end function IsConvInTerminateState */

/**************************************************************************/
/*                                                                        */
/*  FUNCTION: AboutDlgProc                                                */
/*                                                                        */
/*  PURPOSE:  About Dialog box window procedure                           */
/*                                                                        */
/**************************************************************************/

BOOL FAR PASCAL __export AboutDlgProc( HWND   hDlg,
                                       UINT   msg,
                                       WPARAM wParam,
                                       LPARAM lParam )
{

   BOOL  bProcessed = FALSE ;

   switch (msg) {

   case WM_INITDIALOG :

         bProcessed = TRUE ;

   case WM_COMMAND :

     if (wParam == IDOK || wParam == IDCANCEL) {
       EndDialog(hDlg, TRUE) ;

       bProcessed = TRUE ;
     } /* end if */

         break ;

   }  /* end switch */

   return bProcessed ;

}  /* end function AboutDlgProc */

/**************************************************************************/
/*                                                                        */
/*  General Window Output functions                                       */
/*                                                                        */
/**************************************************************************/
/**************************************************************************/
/*                                                                        */
/*  FUNCTION: Info                                                        */
/*                                                                        */
/*  PURPOSE:  Write the text string to the output window                  */
/*                                                                        */
/**************************************************************************/
void Info( char *Format, ... )
{
  va_list ArgPtr ;

  va_start ( ArgPtr, Format ) ;
  vsprintf ( szBuffer, Format, ArgPtr ) ;
  va_end ( ArgPtr ) ;

  SendMessage ( hwndList, LB_INSERTSTRING, (WPARAM)-1, (LPARAM)(LPSTR) szBuffer ) ;

} /* end function Info */

/**************************************************************************/
/*                                                                        */
/*  FUNCTION: Response                                                    */
/*                                                                        */
/*  PURPOSE:  Convert the ECI return code to a text string and display    */
/*                                                                        */
/**************************************************************************/
void Response ( char *Call, short sRetCode )
{
  char *pszError ;

  switch ( sRetCode ) {

    case ECI_NO_ERROR :
      pszError = "ECI_NO_ERROR" ;
      break ;
    case ECI_ERR_INVALID_DATA_LENGTH :
      pszError = "ECI_ERR_INVALID_DATA_LENGTH" ;
      break ;
    case ECI_ERR_INVALID_EXTEND_MODE :
      pszError = "ECI_ERR_INVALID_EXTEND_MODE" ;
      break ;
    case ECI_ERR_NO_CICS :
      pszError = "ECI_ERR_NO_CICS" ;
      break ;
    case ECI_ERR_CICS_DIED :
      pszError = "ECI_ERR_CICS_DIED" ;
      break ;
    case ECI_ERR_REQUEST_TIMEOUT :
      pszError = "ECI_ERR_REQUEST_TIMEOUT" ;
      break ;
    case ECI_ERR_RESPONSE_TIMEOUT :
      pszError = "ECI_ERR_RESPONSE_TIMEOUT" ;
      break ;
    case ECI_ERR_TRANSACTION_ABEND :
      pszError = "ECI_ERR_TRANSACTION_ABEND" ;
      break ;
    case ECI_ERR_EXEC_NOT_RESIDENT :
      pszError = "ECI_ERR_EXEC_NOT_RESIDENT" ;
      break ;
    case ECI_ERR_SYSTEM_ERROR :
      pszError = "ECI_ERR_SYSTEM_ERROR" ;
      break ;
    case ECI_ERR_NULL_WIN_HANDLE :
      pszError = "ECI_ERR_NULL_WIN_HANDLE" ;
      break ;
    case ECI_ERR_NULL_MESSAGE_ID :
      pszError = "ECI_ERR_NULL_MESSAGE_ID" ;
      break ;
    case ECI_ERR_THREAD_CREATE_ERROR :
      pszError = "ECI_ERR_THREAD_CREATE_ERROR" ;
      break ;
    case ECI_ERR_INVALID_CALL_TYPE :
      pszError = "ECI_ERR_INVALID_CALL_TYPE" ;
      break ;
    case ECI_ERR_ALREADY_ACTIVE :
      pszError = "ECI_ERR_ALREADY_ACTIVE" ;
      break ;
    case ECI_ERR_RESOURCE_SHORTAGE :
      pszError = "ECI_ERR_RESOURCE_SHORTAGE" ;
      break ;
    case ECI_ERR_NO_SESSIONS :
      pszError = "ECI_ERR_NO_SESSIONS" ;
      break ;
    case ECI_ERR_NULL_SEM_HANDLE :
      pszError = "ECI_ERR_NULL_SEM_HANDLE" ;
      break ;
    case ECI_ERR_INVALID_DATA_AREA :
      pszError = "ECI_ERR_INVALID_DATA_AREA" ;
      break ;
    default :
      pszError = "!!!Unknown!!!" ;
      break ;
  } /* endswitch */

  Info ( "%s returned: %s", Call, pszError ) ;

  if ( sRetCode == ECI_ERR_SYSTEM_ERROR ) {

    /* Display the system error response: */
    Info ( "SysErr returns: %d(0x%04.4X)",
           eci_parms.eci_sys_return_code,
           eci_parms.eci_sys_return_code ) ;
  } /* endif */

} /* end function Response */

