/*********************************************************************/
/*                                                                   */
/* MODULE NAME        DDESERV.H                                      */
/*                                                                   */
/* DESCRIPTIVE NAME   DDE Server for CICS OS/2 ECI                   */
/*                                                                   */
/* Statement:         Sample Code - No Warranty Provided.            */
/*                                                                   */
/*                    (c) Copyright IBM Corp. 1994.                  */
/*                                                                   */
/*                    All rights reserved.                           */
/*                                                                   */
/* Status:            Version 1 Release 0                            */
/*                                                                   */
/*  NOTES :-                                                         */
/*    DEPENDENCIES = Windows 3.X                                     */
/*    RESTRICTIONS = none                                            */
/*    MODULE TYPE  = Source Header Module                            */
/*    PROCESSOR    = PS/2 and PC                                     */
/*                                                                   */
/*********************************************************************/

#define APPL_NAME             "DDESERV"

#define APP_MAX_SIZE          9    /* application name is DDESERV  */
#define TOPIC_MAX_SIZE        9    /* Topic is always CICS program name */
#define ITEM_NAME_MAX_SIZE    255  /* Item is commarea to be passed. For demo 256 will suffice */
#define ITEM_VALUE_MAX_SIZE   255  /* return value maximum size is same as send size */
#define CONV_MAX_COUNT        4    /* max number of simultaneous conversations */

#define ECI_MENU              "EciTestMenu"
#define ECI_CLASS             "EciTestClass"
#define ECI_ABOUT             "EciAboutDlgBox"
#define DDESERVER_CLASS       "DDEServerWndClass"

/* Control ID's */

#define ECI_MSG               WM_USER + 10
#define ECI_ITEM_EXIT                   256
#define ECI_ITEM_CALL                   257
#define ECI_ITEM_CLEAR                  258
#define ECI_ITEM_ABOUT                  259
#define ECI_ITEM_NULL                   999
#define IDC_STATIC                      -1

#define DEFAULT_ACK_TIME_OUT_MILLISEC 10000

/*********************************************************************/
/* Function Prototypes                                               */
/*********************************************************************/

int  PASCAL WinMain ( HINSTANCE, HINSTANCE, LPSTR, int ) ;

long FAR PASCAL _export MainWindowProc ( HWND, UINT, WPARAM, LPARAM ) ;
long FAR PASCAL _export DDEWindowProc  ( HWND, UINT, WPARAM, LPARAM ) ;
BOOL FAR PASCAL _export AboutDlgProc   ( HWND, UINT, WPARAM, LPARAM ) ;

short ECICall ( char *, HWND ) ;
short ECIRequestData ( void ) ;

void ServerInitiate(HWND, LONG) ;
void ServerRequest(HWND, HWND, LONG) ;
void ServerTerminate(HWND, HWND) ;
void SendData(HWND, HWND, char *, char *, BOOL);
void SendTerminate(HWND , HWND);
void TerminateConversations(void);
void Info ( char *, ... ) ;
void Response ( char *, short ) ;

BOOL AddConv(HWND, HWND ) ;
BOOL AddConvTopic ( HWND, ATOM ) ;
BOOL AddConvItem ( HWND, ATOM ) ;
ATOM GetConvTopic ( HWND ) ;
ATOM GetConvItem ( HWND ) ;
BOOL AtLeastOneConvActive(void) ;
HWND GetHwndDDEClient(HWND) ;
HWND GetNextConv(HWND) ;
BOOL IsConvInTerminateState(HWND) ;
void RemoveConv(HWND) ;
void SetConvInTerminateState(HWND) ;
struct DDECONV * NEAR FindConv(HWND ) ;

/********************************************************************/
/* Global Variables                                                 */
/********************************************************************/

HINSTANCE   hInstance ;              /* Instance handle */
HWND        hwndFrame ;              /* Window Handle for frame   */
HWND        hwndList ;               /* Window Handle for listbox */
HWND        hwndMain ;               /* Window Handle for main window */

char  szTopicName[TOPIC_MAX_SIZE+1] ;      /* to save topic name */
char  szBuffer     [512] ;                 /* for output string to main window */
char  szCommArea   [256] ;                 /* commarea to be passed on ECI call */
char  szItem[ITEM_NAME_MAX_SIZE+1] ;       /* sent item value from client */
char  szItemValue[ITEM_VALUE_MAX_SIZE+1] ; /* returned item value from server */

ECI_PARMS   eci_parms ;                      /* ECI Control block */

typedef struct DDECONV { HWND hwndDDEServer ;
                         HWND hwndDDEClient ;
                         ATOM aClientTopic ;
                         ATOM aClientItem ;
                         BOOL bInClientRequestedTerminate ;
};

struct DDECONV Conv[CONV_MAX_COUNT] ;

static int  nConvCount = 0 ;
static int  nAckTimeOut ;
